/*
  Copyright (c) 2016 Red Hat, Inc. <http://www.redhat.com>
  This file is part of gluster-block.

  This file is licensed to you under your choice of the GNU Lesser
  General Public License, version 3 or any later version (LGPLv3 or
  later), or the GNU General Public License, version 2 (GPLv2), in all
  cases as published by the Free Software Foundation.
*/


# define   _GNU_SOURCE         /* See feature_test_macros(7) */
# include  <stdio.h>
# include  <dirent.h>
# include  <sys/stat.h>

# include "utils.h"
# include "lru.h"
# include "config.h"
# include "version.h"

struct gbConf gbConf = {
  .glfsLruCount = LRU_COUNT_DEF,
  .logLevel = GB_LOG_INFO,
  .logDir = GB_LOGDIR,
  .cliTimeout = CLI_TIMEOUT_DEF
};

const char *argp_program_version = ""                                 \
  PACKAGE_NAME" ("PACKAGE_VERSION")"                                  \
  "\nRepository rev: https://github.com/gluster/gluster-block.git\n"  \
  "Copyright (c) 2016 Red Hat, Inc. <https://redhat.com/>\n"          \
  "gluster-block comes with ABSOLUTELY NO WARRANTY.\n"                \
  "It is licensed to you under your choice of the GNU Lesser\n"       \
  "General Public License, version 3 or any later version (LGPLv3\n"  \
  "or later), or the GNU General Public License, version 2 (GPLv2),\n"\
  "in all cases as published by the Free Software Foundation.";


int
glusterBlockSetLogLevel(unsigned int logLevel)
{
  if (logLevel >= GB_LOG_MAX) {
    MSG(stderr, "unknown LOG-LEVEL: '%d'\n", logLevel);
    return -1;
  }
  LOCK(gbConf.lock);
  gbConf.logLevel = logLevel;
  UNLOCK(gbConf.lock);

  if (gbCtx == GB_CLI_MODE) {
    LOG("cli", GB_LOG_DEBUG,
        "logLevel now is %s", LogLevelLookup[logLevel]);
  } else {
    LOG("mgmt", GB_LOG_CRIT,
        "logLevel now is %s", LogLevelLookup[logLevel]);
  }

  return 0;
}


/* TODO: use gbConf in cli too, for logLevel/LogDir and other future options
int
glusterBlockSetCliTimeout(size_t timeout)
{
  if (timeout < 0) {
    MSG(stderr, "unknown GB_CLI_TIMEOUT: '%zu'\n", timeout);
    return -1;
  }
  LOCK(gbConf.lock);
  gbConf.cliTimeout = timeout;
  UNLOCK(gbConf.lock);

  return 0;
}
*/


int
glusterBlockCLIOptEnumParse(const char *opt)
{
  int i;


  if (!opt) {
    return GB_CLI_OPT_MAX;
  }

  for (i = 0; i < GB_CLI_OPT_MAX; i++) {
    if (!strcmp(opt, gbCliCmdlineOptLookup[i])) {
      return i;
    }
  }

  return i;
}


int
glusterBlockCLICreateOptEnumParse(const char *opt)
{
  int i;


  if (!opt) {
    return GB_CLI_CREATE_OPT_MAX;
  }

  for (i = 0; i < GB_CLI_CREATE_OPT_MAX; i++) {
    if (!strcmp(opt, gbCliCreateOptLookup[i])) {
      return i;
    }
  }

  return i;
}


int
glusterBlockDaemonOptEnumParse(const char *opt)
{
  int i;


  if (!opt) {
    return GB_DAEMON_OPT_MAX;
  }

  for (i = 0; i < GB_DAEMON_OPT_MAX; i++) {
    /* clip '--' from option */
    while (*opt == '-') {
      opt++;
    }
    if (!strcmp(opt, gbDaemonCmdlineOptLookup[i])) {
      return i;
    }
  }

  return i;
}


int
blockLogLevelEnumParse(const char *opt)
{
  int i;


  if (!opt) {
    return GB_LOG_MAX;
  }

  for (i = 0; i < GB_LOG_MAX; i++) {
    if (!strcmp(opt, LogLevelLookup[i])) {
      return i;
    }
  }

  return i;
}


int
blockMetaKeyEnumParse(const char *opt)
{
  int i;


  if (!opt) {
    return GB_METAKEY_MAX;
  }

  for (i = 0; i < GB_METAKEY_MAX; i++) {
    if (!strcmp(opt, MetakeyLookup[i])) {
      return i;
    }
  }

  return i;
}


int
blockMetaStatusEnumParse(const char *opt)
{
  int i;


  if (!opt) {
    return GB_METASTATUS_MAX;
  }

  for (i = 0; i < GB_METASTATUS_MAX; i++) {
    if (!strcmp(opt, MetaStatusLookup[i])) {
      return i;
    }
  }

  return i;
}

int blockRemoteCreateRespEnumParse(const char *opt)
{
  int i;


  if (!opt) {
    return GB_REMOTE_CREATE_RESP_MAX;
  }

  for (i = 0; i < GB_REMOTE_CREATE_RESP_MAX; i++) {
    if (strstr(opt, RemoteCreateRespLookup[i])) {
      return i;
    }
  }

  return i;
}


/* On any failure return, epoch atleast */
void
logTimeNow(char *buf, size_t bufSize)
{
  struct tm tm;
  struct timeval tv;


  if (gettimeofday (&tv, NULL) < 0) {
    goto out;
  }

  if (tv.tv_sec && gmtime_r(&tv.tv_sec, &tm) != NULL) {
    strftime (buf, bufSize, "%Y-%m-%d %H:%M:%S", &tm);
    snprintf (buf + strlen(buf), bufSize - strlen(buf), ".%06ld", tv.tv_usec);
    return;
  }

out:
  snprintf(buf, bufSize, "%lu", (unsigned long)time(NULL));
  return;
}


static bool
glusterBlockLogdirCreate(void)
{
  DIR* dir = opendir(gbConf.logDir);


  if (dir) {
    closedir(dir);
  } else if (errno == ENOENT) {
    if (mkdir(gbConf.logDir, 0640) == -1) {
      fprintf(stderr, "mkdir(%s) failed (%s)", gbConf.logDir, strerror (errno));
      return 0;  /* False */
    }
  } else {
    fprintf(stderr, "opendir(%s) failed (%s)", gbConf.logDir, strerror (errno));
    return 0;  /* False */
  }

  return 1;
}


char *
gbClipoffSensitiveDetails(char *buf)
{
  char *out;
  char *ptr;


  if (!buf) {
    return NULL;
  }

  if (GB_STRDUP(out, buf) < 0) {
    LOG("mgmt", GB_LOG_ERROR,
        "clipoffSensitiveDetails: failed to strdup (%s)", strerror(errno));
    return NULL;
  }

  ptr = strstr(out, "PASSWORD");
  if (ptr) {
    ptr = strchr(ptr, ':');
    /* considering output from plain text, json and json-pretty .. */
    while (ptr && (*ptr != ',' && *ptr != '\n')) {
      /* UUID have alphabets, digits and '-' char */
      if (isalnum(*ptr) || (*ptr == '-')) {
        *ptr = '*';
      }
      ptr++;
    }
  }

  return out;
}


void
gbClipoffSensitiveDetailsAtExec(char **data)
{
  char *cmd = *data;
  char *res = NULL;
  char *line = NULL;
  char *tmp = NULL;
  char *ptr = NULL;
  char *ptr2 = NULL;


  if (!cmd) {
    return;
  }

  /* Filter the password details in two cases
   * 1. when dumping targetcli arguments supplied by gluster-block
   * 2. Output from targetcli
   */
  if (strstr(cmd, "password=") || strstr(cmd, "Parameter password is now")) {
    line = strtok(cmd, "\n");
    while (line) {
      ptr = strstr(line, "password=");
      if (!ptr) {
        ptr = strstr(line, "Parameter password is now");
      }
      if (!ptr) {
        GB_ASPRINTF (&res, "%s\n%s", tmp?tmp:"", line);
        GB_FREE(tmp);
        tmp = res;
      } else {
        ptr2 = strchr(ptr, '=');
        if (!ptr2) {
          ptr2 = strchr(ptr, '\'');
        }
        ptr2++;
        while(ptr2) {
          if (*ptr2 == '\n' || *ptr2 == '\0' || *ptr2 == ' ' || *ptr2 == '\'' || *ptr2 == '.')
            break;
          if (isalnum(*ptr2) || (*ptr2 == '-')) {
            *ptr2 = '*';
          }
          ptr2++;
        }
        GB_ASPRINTF (&res, "%s\n%s", tmp?tmp:"", line);
        GB_FREE(tmp);
        tmp = res;
      }
      line = strtok(NULL, "\n");
    }
    GB_FREE(*data);
    *data = res;
  }
}


int
initLogging(void)
{
  char *logDir = NULL;


  logDir = getenv("GB_LOGDIR");
  if (!logDir) {
    logDir = GB_LOGDIR;
  }

  if (strlen(logDir) > PATH_MAX - GB_MAX_LOGFILENAME) {
    fprintf(stderr, "strlen of logDir Path > PATH_MAX: %s\n", logDir);
    return EXIT_FAILURE;
  }

  /* set logfile paths */
  snprintf(gbConf.logDir, PATH_MAX,
           "%s", logDir);
  snprintf(gbConf.daemonLogFile, PATH_MAX,
           "%s/gluster-blockd.log", logDir);
  snprintf(gbConf.cliLogFile, PATH_MAX,
           "%s/gluster-block-cli.log", logDir);
  snprintf(gbConf.gfapiLogFile, PATH_MAX,
           "%s/gluster-block-gfapi.log", logDir);
  snprintf(gbConf.configShellLogFile, PATH_MAX,
           "%s/gluster-block-configshell.log", logDir);
  snprintf(gbConf.cmdhistoryLogFile, PATH_MAX,
           "%s/cmd_history.log", logDir);

  if(!glusterBlockLogdirCreate()) {
    return EXIT_FAILURE;
  }

  return 0;
}


int
gbRunnerExitStatus(int exitStatus)
{
  if (!WIFEXITED(exitStatus)) {
    return -1;
  }

  return WEXITSTATUS(exitStatus);
}


int
gbRunner(char *cmd)
{
  int childExitStatus;


  childExitStatus = system(cmd);

  return gbRunnerExitStatus(childExitStatus);
}


char*
gbRunnerGetOutput(char *cmd)
{
  FILE *fp;
  char *tptr;
  char *buf = NULL;


  LOG("mgmt", GB_LOG_DEBUG, "command, %s", cmd);

  if (GB_ALLOC_N(buf, 1024) < 0) {
    LOG("mgmt", GB_LOG_ERROR,
        "gbRunnerGetOutput: error allocating memory (%s)", strerror(errno));
    return NULL;
  }

  fp = popen(cmd, "r");
  if (fp) {
    size_t newLen = fread(buf, sizeof(char), 1024, fp);
    if (ferror(fp) != 0) {
      LOG("mgmt", GB_LOG_ERROR,
          "reading command output for %s failed (%s)", cmd, strerror(errno));
      buf[0] = '\0';
      goto fail;
    } else {
      buf[newLen++] = '\0';
      tptr = strchr(buf,'\n');
      if (tptr) {
        *tptr = '\0';
      }
      goto out;
    }
  } else {
    LOG("mgmt", GB_LOG_ERROR,
        "popen(%s): failed: %s", cmd, strerror(errno));
    goto fail;
  }

fail:
  GB_FREE(buf);

 out:
  if (fp) {
    pclose(fp);
  }
  return buf;
}


bool
gbDependencyVersionCompare(int dependencyName, char *version)
{
  size_t vNum[VERNUM_BUFLEN] = {0, };
  char *verStr;
  int i = 0, j;
  char *token, *tmp;
  bool done = false;
  bool ret = false;


  if (GB_STRDUP(verStr, version) < 0) {
    LOG("mgmt", GB_LOG_ERROR,
        "gbDependencyVersionCompare: failed to strdup (%s)", strerror(errno));
    return ret;
  }

  token = strtok(verStr, ".-");
  while( token != NULL ) {
    done = false;
    for (j = 0; j < strlen(token); j++) { /* say if version is 2.1.fb49, parse fb49 */
      if (isdigit(token[j])) {
        vNum[i] = atoi(token);
      } else {
        tmp = token;
        for(; *tmp; ++tmp) {
          if (isdigit(*tmp)) {
            vNum[i] = atoi(tmp);  /* feed 49 from fb49, so done = true */
            done = true;
            break;
          }
        }
      }
      if (done) {
        break;
      }
    }
    token = strtok(NULL, ".-");
    i++;
  }

  switch (dependencyName) {
  case TCMURUNNER:
    if (DEPENDENCY_VERSION(vNum[0], vNum[1], vNum[2]) >= GB_MIN_TCMURUNNER_VERSION_CODE) {
      ret = true;
    }
    break;
  case TARGETCLI:
    if (DEPENDENCY_VERSION(vNum[0], vNum[1], vNum[2]) >= GB_MIN_TARGETCLI_VERSION_CODE) {
      ret = true;
    }
    break;
  case CONFIGSHELL_SEMICOLON:
    if (DEPENDENCY_VERSION(vNum[0], vNum[1], vNum[2]) >= GB_MIN_CONFIGSHELL_SEM_VERSION_CODE) {
      ret = true;
    }
    break;
  }

  GB_FREE(verStr);
  return ret;
}


char *
gbGetRpmPkgVersion(const char* pkgName)
{
  char *exec;

  if (GB_ASPRINTF(&exec, GB_RPM_PKG_VERSION, pkgName) == -1) {
    return NULL;
  }

  return gbRunnerGetOutput(exec);
}


char *
gbRunnerGetPkgVersion(const char * pkgName)
{
  char *cmd = NULL;
  char *out;

  if (!strcmp(pkgName, TARGETCLI_STR)) {
    cmd = TARGETCLI_VERSION;
  } else if (!strcmp(pkgName, RTSLIB_STR)) {
    cmd = RTSLIB_VERSION;
  } else if (!strcmp(pkgName, CONFIGSHELL_STR)) {
    cmd = CONFIGSHELL_VERSION;
  } else if (!strcmp(pkgName,TCMU_STR)) {
    cmd = TCMU_VERSION;
  }

  if (!cmd) {
    return NULL;
  }

  out = gbRunnerGetOutput(cmd);
  if (!out[0] || !strcmp(out, "GIT_VERSION")) {
      out = gbGetRpmPkgVersion(pkgName);
  }

  return out;
}


int
gbAlloc(void *ptrptr, size_t size,
        const char *filename, const char *funcname, size_t linenr)
{
  *(void **)ptrptr = calloc(1, size);

  if (*(void **)ptrptr == NULL) {
    errno = ENOMEM;
    return -1;
  }

  return 0;
}


int
gbAllocN(void *ptrptr, size_t size, size_t count,
         const char *filename, const char *funcname, size_t linenr)
{
  *(void**)ptrptr = calloc(count, size);

  if (*(void**)ptrptr == NULL) {
    errno = ENOMEM;
    return -1;
  }

  return 0;
}


int
gbReallocN(void *ptrptr, size_t size, size_t count,
         const char *filename, const char *funcname, size_t linenr)
{
  void *tmp;


  if (xalloc_oversized(count, size)) {
    errno = ENOMEM;
    return -1;
  }
  tmp = realloc(*(void**)ptrptr, size * count);
  if (!tmp && ((size * count) != 0)) {
    errno = ENOMEM;
    return -1;
  }
  *(void**)ptrptr = tmp;

  return 0;
}


void
gbFree(void *ptrptr)
{
  int save_errno = errno;


  if(*(void**)ptrptr == NULL) {
   return;
  }

  free(*(void**)ptrptr);
  *(void**)ptrptr = NULL;
  errno = save_errno;
}


int
gbStrdup(char **dest, const char *src,
         const char *filename, const char *funcname, size_t linenr)
{
  *dest = NULL;

  if (!src) {
    return 0;
  }

  if (!(*dest = strdup(src))) {
    return -1;
  }

  return 0;
}


char *
gbStrcpy(char *dest, const char *src, size_t destbytes,
         const char *filename, const char *funcname, size_t linenr)
{
    char *ret;
    size_t n = strlen(src);

    if (n > (destbytes - 1))
      return NULL;

    ret = memcpy(dest, src, n);
    dest[n] = '\0';

    return ret;
}


char *
gbStrcat(char *dest, const char *src, size_t destbytes,
         const char *filename, const char *funcname, size_t linenr)
{
    char *ret;
    size_t n = strlen(src);
    size_t m = strlen(dest);

    if (n > (destbytes - 1))
      return NULL;

    ret = memcpy(dest + m, src, n);
    dest[m + n] = '\0';

    return ret;
}
